import urllib.request
import collections
import math
import os
import random
import zipfile
import datetime as dt
import calendar
import time
import pickle

import numpy as np
import tensorflow as tf


######################################################
# Download a file from the internet
#######################################################
def maybe_download(filename, url, expected_bytes):
    """Download a file if not present, and make sure it's the right size."""
    if not os.path.exists(filename):
        filename, _ = urllib.request.urlretrieve(url + filename, filename)
    statinfo = os.stat(filename)
    if statinfo.st_size == expected_bytes:
        print('Found and verified', filename)
    else:
        print(statinfo.st_size)
        raise Exception(
            'Failed to verify ' + filename + '. Can you get to it with a browser?')
    return filename

######################################################
# Read the data into a list of strings.
# filename= zip file containing several files
######################################################
def read_data(filename):
    """Extract the first file enclosed in a zip file as a list of words."""
    with zipfile.ZipFile(filename) as f:
        data = tf.compat.as_str(f.read(f.namelist()[0])).split()
    return data

######################################################
# Create a dictionary, reverse dictionary
# inputs:   words=list of words,
#           n_words= required number of words
# outputs:  count = 2D list of the most common words totaling n_words
#           data = list of word indexes (only those are within if (val < 20)most common words)
#           dictionary = maps the word to an index
#           reverse_dictionary = index to word
######################################################
def build_dataset(words, n_words):
    """Process raw inputs into a dataset."""
    lm_corpus =words[:250]

    count = [['UNK', -1]]
    count.extend(collections.Counter(words).most_common(n_words - 1))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)

    #Add lm_corpus words to dictionary
    for j in range(len(lm_corpus)):
        if not (lm_corpus[j] in dictionary):
            dictionary[lm_corpus[j]]= len(dictionary)

    data = list()
    unk_count = 0
    for word in words:
        if word in dictionary:
            index = dictionary[word]
        else:
            index = 0  # dictionary['UNK']
            unk_count += 1
        data.append(index)
    count[0][1] = unk_count
    reversed_dictionary = dict(zip(dictionary.values(), dictionary.keys()))
    return data, count, dictionary, reversed_dictionary

#########################################################################
# Download, read and build_dataset from the URL http://mattmahoney.net/dc/
##########################################################################
def collect_data(vocabulary_size=10000):
    url = 'http://mattmahoney.net/dc/'
    filename = maybe_download('text8.zip', url, 31344016)
    all_words = read_data(filename)

    data, count, dictionary, reverse_dictionary = build_dataset(all_words,
                                                                vocabulary_size)
    lm_corpus=all_words[:250]
    #
    del all_words  # Hint to reduce memory.
    return lm_corpus, data, count, dictionary, reverse_dictionary



data_index = 0
#########################################################################
# generate batch data. Returns a batch of data given the data
#########################################################################
def generate_batch(data, batch_size, num_skips, skip_window):
    global data_index
    assert batch_size % num_skips == 0
    assert num_skips <= 2 * skip_window
    batch = np.ndarray(shape=(batch_size), dtype=np.int32)
    context = np.ndarray(shape=(batch_size, 1), dtype=np.int32)
    span = 2 * skip_window + 1  # [ skip_window input_word skip_window ]
    buffer = collections.deque(maxlen=span)
    for _ in range(span):
        buffer.append(data[data_index])
        data_index = (data_index + 1) % len(data)
    for i in range(batch_size // num_skips):
        target = skip_window  # input word at the center of the buffer
        targets_to_avoid = [skip_window]
        for j in range(num_skips):
            while target in targets_to_avoid:
                target = random.randint(0, span - 1)
            targets_to_avoid.append(target)
            batch[i * num_skips + j] = buffer[skip_window]  # this is the input word
            context[i * num_skips + j, 0] = buffer[target]  # these are the context words
        buffer.append(data[data_index])
        data_index = (data_index + 1) % len(data)
    # Backtrack a little bit to avoid skipping words in the end of a batch
    data_index = (data_index + len(data) - span) % len(data)
    return batch, context

#########################################################################
# Script execution starts here
#########################################################################
vocabulary_size_ = 10000
lm_corpus, data, count, dictionary, reverse_dictionary = collect_data(vocabulary_size=vocabulary_size_)

vocabulary_size=len(dictionary)

################################
# Print first 20 entries of the dictionary
#
#
#
###############################

batch_size = 128
embedding_size = 300  # Dimension of the embedding vector.
skip_window = 2       # How many words to consider left and right.
num_skips = 4        # How many times to reuse an input to generate a label.

# We pick a random validation set to sample nearest neighbors. Here we limit the
# validation samples to the words that have a low numeric ID, which by
# construction are also the most frequent.
valid_size = 16     # Random set of words to evaluate similarity on.
valid_window = 100  # Only pick dev samples in the head of the distribution.
valid_examples = np.random.choice(valid_window, valid_size, replace=False)
num_sampled = 64    # Number of negative examples to sample.

graph = tf.Graph()


########################################################################
# Specify the computational graph for the softmax based embedding network
#########################################################################
with graph.as_default():

  # Input data.
  train_inputs = tf.placeholder(tf.int32, shape=[batch_size])
  train_context = tf.placeholder(tf.int32, shape=[batch_size, 1])
  valid_dataset = tf.constant(valid_examples, dtype=tf.int32)

  # Look up embeddings for inputs.
  embeddings = tf.Variable(
      tf.random_uniform([vocabulary_size, embedding_size], -1.0, 1.0))
  embed = tf.nn.embedding_lookup(embeddings, train_inputs)

  # Construct the variables for the softmax
  weights = tf.Variable(
      tf.truncated_normal([embedding_size, vocabulary_size],
                          stddev=1.0 / math.sqrt(embedding_size)))
  biases = tf.Variable(tf.zeros([vocabulary_size]))
  hidden_out = tf.transpose(tf.matmul(tf.transpose(weights), tf.transpose(embed))) + biases

  # convert train_context to a one-hot format
  train_one_hot = tf.one_hot(train_context, vocabulary_size)

  cross_entropy = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=hidden_out, labels=train_one_hot))

  # Construct the SGD optimizer using a learning rate of 1.0.
  optimizer = tf.train.GradientDescentOptimizer(1.0).minimize(cross_entropy)

  # Compute the cosine similarity between minibatch examples and all embeddings.
  norm = tf.sqrt(tf.reduce_sum(tf.square(embeddings), 1, keep_dims=True))
  normalized_embeddings = embeddings / norm
  valid_embeddings = tf.nn.embedding_lookup(
      normalized_embeddings, valid_dataset)
  similarity = tf.matmul(
      valid_embeddings, normalized_embeddings, transpose_b=True)

  # Add variable initializer.
  init = tf.global_variables_initializer()

########################################################################
# Function for running a specified graph
#########################################################################
def run(graph, num_steps):
    with tf.Session(graph=graph) as session:
      # We must initialize all variables before we use them.
      init.run()
      print('Initialized')

      average_loss = 0
      for step in range(num_steps):
        batch_inputs, batch_context = generate_batch(data,batch_size, num_skips, skip_window)

        ##########################################################################
        # for step =0,  print the input batch and target batch of words
        #if (step == 0):
        #
        #
        #
        ###################################################################

        feed_dict = {train_inputs: batch_inputs, train_context: batch_context}


        # We perform one update step by evaluating the optimizer op (including it
        # in the list of returned values for session.run()
        _, loss_val = session.run([optimizer, cross_entropy], feed_dict=feed_dict)
        average_loss += loss_val

        if step % 2000 == 0:
          if step > 0:
            average_loss /= 2000
          # The average loss is an estimate of the loss over the last 2000 batches.
          print('Average loss at step ', step, ': ', average_loss)
          average_loss = 0

        # Note that this is expensive (~20% slowdown if computed every 500 steps)
        if step % 10000 == 0:
          sim = similarity.eval()
          for i in range(valid_size):
            valid_word = reverse_dictionary[valid_examples[i]]
            top_k = 8  # number of nearest neighbors
            nearest = (-sim[i, :]).argsort()[1:top_k + 1]
            log_str = 'Nearest to %s:' % valid_word
            for k in range(top_k):
              close_word = reverse_dictionary[nearest[k]]
              log_str = '%s %s,' % (log_str, close_word)
            print(log_str)
      final_embeddings = normalized_embeddings.eval()
    return final_embeddings

#############################################################
# Save corpus necessary for language modelling with RNN
############################################################
def save_lm_corpus(final_embeddings):
      #Save language model corpus for RNN  task_
      lm_dict=dict()
      lm_embeddings =[]
      for k in range(len(lm_corpus)):
          cur_embedding= final_embeddings[dictionary[lm_corpus[k]]]
          #print(lm_corpus[k], dictionary[lm_corpus[k]],cur_embedding[:5] )
          if not (lm_corpus[k] in lm_dict):
              lm_dict[lm_corpus[k]] =len(lm_dict)
              lm_embeddings.append(cur_embedding)

      with open('lm_dict.pkl', 'wb') as f:
          pickle.dump(lm_dict, f, pickle.HIGHEST_PROTOCOL)

      with open('lm_embeddings.pkl', 'wb') as f:
          pickle.dump(lm_embeddings, f, pickle.HIGHEST_PROTOCOL)

      with open('lm_corpus.pkl', 'wb') as f:
          pickle.dump(lm_corpus, f, pickle.HIGHEST_PROTOCOL)

########################################################################
# Run the softmax based graph
#########################################################################
num_steps = 100

print ("\n##########################\n")
print("Starting softmax method")
print ("\n##########################\n")

softmax_start_time = dt.datetime.now()
fin_emb=run(graph, num_steps=num_steps)
softmax_end_time = dt.datetime.now()



print ("\n##########################\n")
print("Softmax method took {} seconds to run 100 iterations".format((softmax_end_time-softmax_start_time).total_seconds()))
print ("\n##########################\n")


########################################################################
# Specify the computational graph for Noise Contrastive Estimation (NCE)
#########################################################################
with graph.as_default():

    # Construct the variables for the NCE loss
    nce_weights = tf.Variable(
        tf.truncated_normal([vocabulary_size, embedding_size],
                            stddev=1.0 / math.sqrt(embedding_size)))
    nce_biases = tf.Variable(tf.zeros([vocabulary_size]))

    nce_loss = tf.reduce_mean(
        tf.nn.nce_loss(weights=nce_weights,
                       biases=nce_biases,
                       labels=train_context,
                       inputs=embed,
                       num_sampled=num_sampled,
                       num_classes=vocabulary_size))

    optimizer = tf.train.GradientDescentOptimizer(1.0).minimize(nce_loss)

    # Add variable initializer.
    init = tf.global_variables_initializer()


########################################################################
# Run the NCE graph
#########################################################################
num_steps = 50000

print ("\n##########################\n")
print("Starting NCE method")
print ("\n##########################\n")

nce_start_time = dt.datetime.now()
fin_emb_nce=run(graph, num_steps)
nce_end_time = dt.datetime.now()

save_lm_corpus(fin_emb_nce)

print ("\n##########################\n")
print("NCE method took {} seconds to run 100 iterations".format((nce_end_time-nce_start_time).total_seconds()*0.002))
print ("\n##########################\n")